//------------------------------------------------------------------------
//
// DCC-Polarity
//
// Copyright (c) 2012-2013 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
// 
//------------------------------------------------------------------------
//
// file:      DCC-Polarity.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2013-11-22 V0.01 start
//            2013-12-22 V1.00 release
//            2014-04-06 V1.01 LED fr DC umgetauscht
//
// tests:     AC, DC, DCC done
//            
//------------------------------------------------------------------------
//
// purpose:   Show Polarity of DCC
//
// target:    AVR Attiny2313, running @8MHz internal clock.
//          
// content:    1. Defines and variable definitions
//             2. RTOS Helper Functions
//             3. Init
//             4. ISR (only DCC)
//             5. Tools (timeout, blinking)
//             6. MAIN: check the states, trace and retry enable
//
//------------------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/io.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <util/atomic.h>
#include <string.h>
#include <avr/sleep.h>

// Fuses attiny2313:
// Extended:  0xFF
// High:      0xDB
// Low:       0xC4

//==============================================================================
//
// Section 1   Defines
//
//==============================================================================

#define TEST      1         // any Number
#define STANDARD  2

#define SIMULATION  0            // 0: real application
                                 // 1: test receive routine
                                 // 2: test timing engine (toggle fast)
                                 // 3: test motor

#define LED_MODE    STANDARD     // STANDARD 
                                 // TEST use for debugging

#define DCC_RAILCOM  1           // 1: receive of railcom supported

//---------------------------------------------------------------------
// Timing Definitions: (all values in us)
//
#ifndef F_CPU
   // prevent compiler error by supplying a default 
   # warning "F_CPU not defined for <booster_timer.c>", set default to 8MHz 
   # define F_CPU 8000000UL
   // if changed: check every place where it is used
   // (possible range underflow/overflow in preprocessor!)
#endif
#if (F_CPU > 30000000)
  #warning possible long int overflow - please check all define calculations
#endif
//---------------------------------------------------------------------
//
//---------------------------------------------------------------------
// Definition for DCC-Timing
/// This are DCC timing definitions from NMRA
#define PERIOD_1   116L          // 116us for DCC 1 pulse - do not change
#define PERIOD_0   232L          // 232us for DCC 0 pulse - do not change
#define MAX_DCC_SIZE  16

#define RAILCOM_CUTOUT_START   (30L)      // min. 30us from Endbit until Datagram1
#define ISR_LATENCY             13
#define RAILCOM_CUTOUT_END     (454L)     // min. 454us from Endbit until end of Datagram2
                                          // sub. 13: latency from ISR
#define RAILCOM_DEADTIME    4                                            // 4us take over

//---------------------------------------------------------------------
// Definition for OUTPUT and Timing
// normal timer mode: 8MHz / 8 / 65536 => 65ms / Tick
// CTC, Overflow at 0x0FFF:            =>  4ms / Tick

#define TICK_PERIOD                    4096L    // 65ms tick for Timer

#define BLINK_PERIOD                 600000L

#define SECOND_PERIOD               1000000L

#define AC_TIMEOUT_PERIOD            100000L    // wenn 100ms keine Flanke


//---------------------------------------------------------------------------
// PORT Definitions:
//
// PORTD:
#define PROGTASTER    0     // input, internal pullup
#define DCC_IN        2     // must be located on INT0
#define LED_RED1      5     // out
#define LED_BLUE1     4     // out


// PORTB:
#define LED_RED2      0    // out
#define LED_BLUE2     1     // out
#define NU_PB2        2

#define PROG_PRESSED   (!(PIND & (1<<PROGTASTER)))
#define JUMPER_FITTED  (!(PIND & (1<<JUMPER)))
#define LED_RED1_OFF    PORTD |= (1<<LED_RED1)
#define LED_RED1_ON     PORTD &= ~(1<<LED_RED1)
#define LED_BLUE1_OFF    PORTD |= (1<<LED_BLUE1)
#define LED_BLUE1_ON     PORTD &= ~(1<<LED_BLUE1)

#define LED_RED2_OFF    PORTB |= (1<<LED_RED2)
#define LED_RED2_ON     PORTB &= ~(1<<LED_RED2)
#define LED_BLUE2_OFF    PORTB |= (1<<LED_BLUE2)
#define LED_BLUE2_ON     PORTB &= ~(1<<LED_BLUE2)



#define DCC_IN_HIGH        ( (PIND & (1<<DCC_IN)))
#define DCC_IN_LOW         (!(PIND & (1<<DCC_IN)))
#define DCCIN_STATE        (  PIND & (1<<DCC_IN) )

#define LED_DCC_ON          PORTB |= (1<<NU_PB2)            // Pin 13
#define LED_DCC_OFF         PORTB &= ~(1<<NU_PB2)

//----------------------------------------------------------------------------
// Internal Hardware usage:
//
// Timer 1: prescaler 8, running @ 1MHz, normal mode. 
//          permanently running, wrap around; 
//          OCRx used to create cutout
//
//----------------------------------------------------------------------------
// Global Data

#define OPTIMIZE_VARS 1

#if (OPTIMIZE_VARS == 1)
     #define Communicate GPIOR0
     #define Recstate GPIOR1
     #define DCCstate GPIOR2
#else
     unsigned char Communicate;
     unsigned char Recstate; 
     unsigned char DCCstate;
#endif

// timer access
typedef union { unsigned int ui16;
                signed int i16;
                struct {  signed int i10: 10;
                          unsigned char rest: 4;
                       };
                struct {  signed char  i8l;
                          signed char  i8h;
                       };
                struct {  unsigned char  ui8l;
                          unsigned char  ui8h;
                       };
              } t_timer;

// t_timer last_dcc_edge;

signed char last_dcc_edge;
signed char last_dcc_run;

// signed int period_1_4x = PERIOD_1 * 4;
// signed int period_1_1x = PERIOD_1;
// t_timer railcom_refpoint;      // loaded by DCC Tracker

const unsigned char version[] PROGMEM    = {" -- DCC Polarity Tester, V1.0"};
const unsigned char copyright1[] PROGMEM = {" -- (c) Kufer Wolfgang, 22.12.2013 "};
const unsigned char copyright2[] PROGMEM = {" -- www.opendcc.de -- "};

typedef enum
  {
    AC,
    DC_P,
    DC_N,
    DCC_P,
    DCC_N,
  } t_in_state;

t_in_state in_state;

//----------------------------------------------------------------------------------
// Definitions for inter process communication


#define C_DCC_Received    0    // a new DCC was received - issued by ISR(Timer1)
                               //                          cleared by main
#define C_Railcom         1    // Railcom window is open
                               //                        - issued by ISR
                               //                          cleared by watch_railcom
#define C_RailcomRec      2    // we are receiving a cutout

#define C_DCC_Edge        3    // a new DCC edge was received - issued by ISR(INT0)
                               //                          cleared by clock adjust



//==============================================================================
//
// Section 2   RTOS Helper Functions
//
//==============================================================================


static inline unsigned char semaphor_query(unsigned char flag) 
       __attribute__((always_inline));

unsigned char semaphor_query(unsigned char flag)
  {
    return (Communicate & (1<<flag));
  }
               
static inline void semaphor_set(unsigned char flag) 
       __attribute__((always_inline));

void semaphor_set(unsigned char flag)
  {
    cli();
    Communicate |= (1<<flag);
    sei();
  }

static inline void semaphor_set_isr(unsigned char flag) 
       __attribute__((always_inline));

void semaphor_set_isr(unsigned char flag)
  {
    Communicate |= (1<<flag);
  }
       
static inline unsigned char semaphor_get(unsigned char flag) 
       __attribute__((always_inline));

unsigned char semaphor_get(unsigned char flag)
  {
    unsigned char value;
    cli();
    value = Communicate & (1<<flag);
    Communicate &= (~(1<<flag));
    sei();
    return(value);
  }

static inline void semaphor_clear(unsigned char flag) 
       __attribute__((always_inline));

void semaphor_clear(unsigned char flag)
  {
    cli();
    Communicate &= (~(1<<flag));
    sei();
  }

//==============================================================================
//
// Section 3   INIT
//
//==============================================================================


static void init_main(void)
  {
    DDRB  = (1<<LED_RED2)  
          | (1<<LED_BLUE2) ; 


    PORTB  = (1<<LED_RED2)  
           | (1<<LED_BLUE2) ;    // both 1 = off
          

    DDRD  = (0<<PROGTASTER)     // in - set to 0
          | (0<<DCC_IN)  
          | (1<<LED_RED1)  
          | (1<<LED_BLUE1) ; 
    PORTD = (1<<PROGTASTER)     // pullup - set to 1
          | (0<<DCC_IN)         // DCC IN = pull down
          | (1<<LED_RED1)  
          | (1<<LED_BLUE1) ;    // both 1 = off


    // LEDs are all on

    // Init Timer0

    #define T0_PRESCALER   8    // may be 1, 8, 64, 256, 1024
    #if   (T0_PRESCALER==1)
        #define T0_PRESCALER_BITS   ((0<<CS02)|(0<<CS01)|(1<<CS00))
    #elif (T0_PRESCALER==8)
        #define T0_PRESCALER_BITS   ((0<<CS02)|(1<<CS01)|(0<<CS00))
    #elif (T0_PRESCALER==64)
        #define T0_PRESCALER_BITS   ((0<<CS02)|(1<<CS01)|(1<<CS00))
    #elif (T0_PRESCALER==256)
        #define T0_PRESCALER_BITS   ((1<<CS02)|(0<<CS01)|(0<<CS00))
    #elif (T0_PRESCALER==1024)
        #define T0_PRESCALER_BITS   ((1<<CS02)|(0<<CS01)|(1<<CS00))
    #endif

//    #define T87US (F_CPU * PERIOD_1 * 3 / 4 / T0_PRESCALER / 1000000L)

    #define T87US (F_CPU * 77L / T0_PRESCALER / 1000000L)

    #if (T87US > 254)
      #warning T87US too big, use either larger prescaler or slower processor
    #endif
    #if (T87US < 32)
      #warning T87US too small, use either smaller prescaler or faster processor
    #endif

    TCCR0A  = (0 << COM0A1)
            | (0 << COM0A0)       // com = 00: normal mode, pin operates as usual
            | (0 << COM0B1) 
            | (0 << COM0B0)       // com = 00: normal mode, pin operates as usual
            | (0 << WGM01)        // wgm = 00: normal mode, top=0xff       
            | (0 << WGM00);       // wgm = 00: normal mode, top=0xff
          
    TCCR0B  =  (0 << FOC0A)       // force output compare
             | (0 << FOC0B)       
             | (0 << WGM02)       
             | 0;                 // stopped (T0_PRESCALER_BITS);    //   = run 

    TCNT0 = 256L - T87US;  

    // OCR0 is unused -> could be used for flags!

    
    // Init Timer1, tick 1us
    // Overflow at 0x0FFF (Mode 12)
    
    #define TIMER1_CLKDIV 8              // possible values: 1, 8, 64, 256, 1024

    #define TIMER1_MODE   4             //  4 = CTC, Top = OCR1A
                                        // 12 = CTC, Top = ICR1

    OCR1A = 0x0FFF;

    TCCR1A = (0 << COM1A1)      // compare match A
           | (0 << COM1A0) 
           | (0 << COM1B1)      // compare match B
           | (0 << COM1B0) 
           | 0                  // reserved
           | 0                  // reserved
           | (((TIMER1_MODE >> 1) & 1) << WGM11)
           | (((TIMER1_MODE >> 0) & 1) << WGM10);
    TCCR1B = (0 << FOC1A) 
           | (0 << FOC1B) 
           | (((TIMER1_MODE >> 2) & 1) << WGM12) 
           | (((TIMER1_MODE >> 3) & 1) << WGM13) 
    #if   TIMER1_CLKDIV == 1
           | (0 << CS12) | (0 << CS11)  | (1 << CS10);       // CS[2:0]=001 clkdiv 1
    #elif TIMER1_CLKDIV == 8
           | (0 << CS12) | (1 << CS11)  | (0 << CS10);       // CS[2:0]=010 clkdiv 8
    #elif TIMER1_CLKDIV == 64
           | (0 << CS12) | (1 << CS11)  | (1 << CS10);       // CS[2:0]=011 clkdiv 64 
    #elif TIMER1_CLKDIV == 256
           | (1 << CS12) | (0 << CS11)  | (0 << CS10);       // CS[2:0]=100 clkdiv 256
    #elif TIMER1_CLKDIV == 1024
           | (1 << CS12) | (0 << CS11)  | (1 << CS10);       // CS[2:0]=101 clkdiv 1024   
    #else
     #warning: TIMER1_CLKDIV is void
    #endif


    TIMSK = (0<<TOIE1)          // Timer1 Overflow  (Tick)
          | (1<<OCIE1A)         // Timer1 Compare A
          | (0<<OCIE1B)         // Timer1 Compare B
          | 0                   // reserved
          | (0<<ICIE1)          // Timer1 Input Capture
          | (0<<OCIE0B)         // Timer0 Compare B
          | (1<<TOIE0)          // Timer0 Overflow
          | (0<<OCIE0A);        // Timer0 Compare A

    // Init Interrupt 0

    GIMSK = (0<<INT1)           // Int1 is msb
          | (1<<INT0)           // 1= Enable INT0 -> DCC
          | (0<<PCIE); 

    MCUCR |= (1<<ISC01)         //  1,1: rising; 1,0 falling
           | (1<<ISC00);        //  The rising edge of INT0 generates an interrupt request.
                                //  This may change during run.
                                
    semaphor_clear(C_Railcom);    // clear any railcom window

    TCNT1 = 0;
    last_dcc_edge = 0;
    last_dcc_run = 0;
  }


//==============================================================================
//
// Section 4 Interrupts
//
//==============================================================================



// ISR(INT0) loads only a register and stores this register to IO.
// this influences no status flags in SREG.
// therefore we define a naked version of the ISR with
// no compiler overhead.

#define ISR_INT0_OPTIMIZED  1

#if (ISR_INT0_OPTIMIZED == 1)
    #define ISR_X_NAKED(vector) \
        void vector (void) __attribute__ ((signal, naked)); \
        void vector (void)
    

    ISR_X_NAKED(INT0_vect) 
      {
         asm volatile ( "push r16" ); 
         __asm__ __volatile 
          (
            "ldi r16, %1"  "\n\t"
            "out %0, r16" "\n\t"
         :                          // no output section
         : "M" (_SFR_IO_ADDR (TCCR0B)),
           "M" ( (0 << FOC0A)
               | (0 << FOC0B)
               | (0 << WGM02)       // 
               | (T0_PRESCALER_BITS)       // cs02.01.00
               )
          );
        #if (DCC_RAILCOM == 1)      // save tcnt1l to last_dcc_edge
        __asm__ __volatile 
          (
            "in r16, %1"  "\n\t"
            "sts %0, r16" "\n\t"
         : "=m" ( last_dcc_edge )                         // output section
         : "M" (_SFR_IO_ADDR (TCNT1L))
          );
        #endif
        asm volatile ( "pop r16" ); 
        semaphor_set_isr(C_DCC_Edge);
        asm volatile ( "reti" ); 
      }
#else
    ISR(INT0_vect) 
      {
        TCCR0B = (0 << FOC0A)
               | (0 << FOC0B)
               | (0 << WGM02)       // 
               | (T0_PRESCALER_BITS);    //   = run 

        #if (LED_MODE == TEST)
            #warning LED_MODE = TEST
            LED_RED1_ON; _delay_us(1); LED_RED1_OFF;
        #endif
        #if (DCC_RAILCOM == 1)
            // TCNT2 = 0;                  // reset railcom timer
            last_dcc_edge = TCNT1L;       // save actual edge time for railcom answer
        #endif     
        semaphor_set_isr(C_DCC_Edge);            
      }

#endif

#if (1==0)
ISR(INT0_vect) 
  {
    dcc_edge.ui16 = TCNT1;                  // save timing reference point (16 bit)
    semaphor_set_isr(C_DCC_Edge);       
  }
#endif

//-------------------------------------------------------------------------------
//
// DCC Receive Routine
//
// Howto:    Edge on DCC generates an Int and dumps the timing point to last_dcc_edge.
//           The main program grab this value and check the timing difference to the
//           previous edge.
//
//                           |<-----116us----->|
//
//           DCC 1: _________XXXXXXXXX_________XXXXXXXXX_________
//                           ^-INT0
//                           |--58us--|
//                                    ^compare
//
//           DCC 0: _________XXXXXXXXXXXXXXXXXX__________________
//                           ^-INT0
//                           |-----116us----->|
//                                            ^compare
//           
// Result:   1. The received message is collected in the struct "local"
//           2. After receiving a complete message, data is copied to
//              "incoming".
//           3. The flag C_Received is set.
//
//
// Polarity Detection:
//           after preamble, we look at the time gap for the first 0
//
//           Data:   |----1----|----1----|---------0---------|
//           Positiv: XXXXX_____XXXXX_____XXXXXXXXXX__________XXXXXXXXXX
//           Int:           ^         ^ gap=1,0  ^     gap=2         ^
//
//           Negativ: _____XXXXX_____XXXXX__________XXXXXXXXXX__________
//           Int:       ^         ^          ^  gap=1,5    ^
//
//
// Railcom Implementation
//           1. ensure correct polarity (means: invert polarity if neccessary)
//           2. reset timer2 every rising int
//           3. at endbit of message: set flag -> C_railcom
//           4. in main: check flag and correct time gap -> and send answer
//

#define T_GAP ((F_CPU / TIMER1_CLKDIV * PERIOD_1 * 5 / 4 / 1000000L)%256)  // 1.25 * PERIOD_1

// here just a repetition of the defines in dcc_receiver.h
// #define MAX_DCC_SIZE  6
// typedef struct
//   {
//     unsigned char size;               // 3 .. 6, including XOR
//     unsigned char dcc[MAX_DCC_SIZE];  // the dcc content
//   } t_message;

//!! t_message incoming;

//!! volatile t_message local;


struct
    {
        unsigned char state;                    // current state
        unsigned char bitcount;                 // current bit
        unsigned char bytecount;                // pointer to current byte
        unsigned char accubyte;                 // location for bit stuffing
        signed char dcc_time;                   // integration time for dcc (only sampling code)
                                                // we start with -7 -> all values >= indicate a zero
        unsigned char filter_data;              // bitfield for low pass data
    } dccrec;

// some states:
#define RECSTAT_WF_PREAMBLE  0
#define RECSTAT_WF_RAILCOM   1                  // check for cutout
#define RECSTAT_WF_PREAMBLE2 2
#define RECSTAT_WF_LEAD0     3
#define RECSTAT_WF_BYTE      4
#define RECSTAT_WF_TRAILER   5
#define RECSTAT_WF_SECOND_H  6                  // nur bei der neuen!!!


#define RECSTAT_DCC          7   


// ISR(INT0) loads only a register and stores this register to IO.
// this influences no status flags in SREG.
// therefore we define a naked version of the ISR with
// no compiler overhead.



ISR(TIMER0_OVF_vect)
  {
    #if (LED_MODE == TEST)
        #warning LED_MODE = TEST
        LED_BLUE1_ON; _delay_us(1); LED_BLUE1_OFF;
    #endif
    #define mydcc (Recstate & (1<<RECSTAT_DCC))

    // read asap to keep timing!
    if (MCUCR & (1<<ISC00))
      {
        if (DCCIN_STATE) Recstate &= ~(1<<RECSTAT_DCC);  // if high -> mydcc=0
        else             Recstate |= 1<<RECSTAT_DCC;    
      }
    else
      {
        if (DCCIN_STATE) Recstate |= 1<<RECSTAT_DCC;    // if low -> mydcc=0
        else             Recstate &= ~(1<<RECSTAT_DCC);  
      }    

    TCCR0B = (0 << FOC0A)
           | (0 << FOC0B)
           | (0 << WGM02)       // 
           | (0 << CS02)        // cs02.01.00
           | (0 << CS01)        //    0  0  0 = stop
           | (0 << CS00);       //    0  0  1 = run 1:1

    // Interrupt occurs at MAX+1 (=256)
    // set Timer Value to 256 - (3/4 of period of a one) -> this is a time window of 116*0,75=87us
    
    TCNT0 = 256L - T87US;  

    dccrec.bitcount++;
        
    if (Recstate & (1<<RECSTAT_WF_PREAMBLE))            // wait for preamble
      {                                       
        if (mydcc)
          {
            semaphor_get(C_Railcom);                    // we got a one - kill any existing railcom window
            if (dccrec.bitcount >= 10) 
              {
                Recstate = 1<<RECSTAT_WF_LEAD0; 
                last_dcc_run = TCNT1L;               // save time stamp
              }
          }
        else
          {
            dccrec.bitcount=0;
          }
      }

    else if (Recstate & (1<<RECSTAT_WF_LEAD0))          // wait for leading 0
      {
      #if (DCC_RAILCOM == 1)
        if (mydcc)
          {                                             // still 1, wait again
            last_dcc_run = TCNT1L;                      // save time stamp
          }
        else
          {
            // determine polarity
            // if we have a gap of 116us, we see this 0-Bit in the first half of the cell
            //                            and we are synced with DCC
            // if we have a gap of 174us, we see this 0-Bit in the second half of the cell.
            //                            and we are not synced - so we may need to invert 
            // so we calculate the gap and compare it to the middle
            // middle is PERIOD_1 * 1.25
            // (all in the same modulo domain as the timer0)
            int8_t gap;

            gap = TCNT1L - last_dcc_run;
            gap = gap - T_GAP;
            if (gap >= 0)
              {
                  if (MCUCR & (1<<ISC00)) MCUCR &= ~(1<<ISC00);    //  we need to invert the active edge of INT0.
                  else MCUCR |= (1<<ISC00);
                  Recstate = 1<<RECSTAT_WF_PREAMBLE;
                  // LED_RED2_ON; _delay_us(1); LED_RED2_OFF;
              }
            else
              {
                // LED_RED2_OFF;
                dccrec.bytecount=0;
                Recstate = 1<<RECSTAT_WF_BYTE;
                dccrec.bitcount=0;
                dccrec.accubyte=0;
              }
          }

      #else // DCC_RAILCOM
        if (mydcc)
          {                                             // still 1, wait again
          }
        else
          {
            dccrec.bytecount=0;
            Recstate = 1<<RECSTAT_WF_BYTE;
            dccrec.bitcount=0;
            dccrec.accubyte=0;
          }
      #endif // DCC_RAILCOM  
      }
    else if (Recstate & (1<<RECSTAT_WF_BYTE))           // wait for byte
      {
        unsigned char my_accubyte;
        my_accubyte = dccrec.accubyte << 1;
        if (mydcc)
          {
            my_accubyte |= 1;
          }
        dccrec.accubyte = my_accubyte;
        
        /* dccrec.accubyte = dccrec.accubyte << 1;
        if (mydcc)
          {
            dccrec.accubyte |= 1;
          }
         */
        if (dccrec.bitcount==8)
          {
            if (dccrec.bytecount == MAX_DCC_SIZE)       // too many bytes
              {                                         // ignore message
                Recstate = 1<<RECSTAT_WF_PREAMBLE;   
              }
            else
              {
                // local.dcc[dccrec.bytecount++] = dccrec.accubyte;
                Recstate = 1<<RECSTAT_WF_TRAILER; 
              }
          }
      }
    else if (Recstate & (1<<RECSTAT_WF_TRAILER))        // wait for 0 (next byte) 
      {                                                 // or 1 (eof message)
        if (mydcc)
          {  // trailing "1" received
            Recstate = 1<<RECSTAT_WF_PREAMBLE;
            dccrec.bitcount=1;

            if (semaphor_query(C_DCC_Received))
              {
                // panic - nobody is reading the messages :-((
              }
            else
              {                                         // copy from local to global
                // unsigned char i;
                // for (i=0; i<MAX_DCC_SIZE; i++)
                //   {
                //      incoming.dcc[i] = local.dcc[i];
                //   }
                // incoming.size = dccrec.bytecount;
                semaphor_set(C_DCC_Received);           // ---> tell the main prog!

                #if (DCC_RAILCOM == 1)
                last_dcc_run = TCNT1L;    // save system timer to calculate the gap.
                                          // do it here again to reset time
                semaphor_set(C_Railcom); 
                #endif

              }
            
          }
        else
          {
            Recstate = 1<<RECSTAT_WF_BYTE;
            dccrec.bitcount=0;
            dccrec.accubyte=0;
          }
      }
    else
      {
        Recstate = 1<<RECSTAT_WF_PREAMBLE;
      }
  }
  


//==============================================================================
//
// Section 5 Tools
//
//==============================================================================

volatile signed char timerval;   // Tick 15ms, Range: 2s

ISR(TIMER1_COMPA_vect)   // 
  {
    timerval++;
    // semaphor_set_isr(C_Tick);       
  }

static signed char get_timerval(void) __attribute__((always_inline));
static signed char get_timerval(void)
  {
    signed char retval;
    // ATOMIC_BLOCK(ATOMIC_FORCEON)
     {
       cli();
       retval = timerval;
       sei(); 
     }
    return(retval);
  }

// power-up blinking
void blink_led(void)
  {
    volatile signed char led_time;
    unsigned char i;
    
    for (i=0; i<3; i++)
      {
        LED_RED1_ON; LED_RED2_ON;
        led_time = get_timerval();
        while ((signed char)(get_timerval() - led_time) < (BLINK_PERIOD / 2L / TICK_PERIOD)) {}
        LED_RED1_OFF; LED_RED2_OFF;
        led_time = get_timerval();
        while ((signed char)(get_timerval() - led_time) < (BLINK_PERIOD / 2L / TICK_PERIOD)) {}
        LED_BLUE1_ON; LED_BLUE2_ON;
        led_time = get_timerval();
        while ((signed char)(get_timerval() - led_time) < (BLINK_PERIOD / 2L / TICK_PERIOD)) {}
        LED_BLUE1_OFF; LED_BLUE2_OFF;
        led_time = get_timerval();
        while ((signed char)(get_timerval() - led_time) < (BLINK_PERIOD / 2L / TICK_PERIOD)) {}
      }    
  }

// test blinking 1s
void blink_red(void)
  {
    volatile signed char led_time;

    while(1)
      {
        LED_RED1_ON; LED_RED2_ON;
        led_time = get_timerval();
        while ((signed char)(get_timerval() - led_time) < (SECOND_PERIOD / 2L / TICK_PERIOD)) {}
        LED_RED1_OFF; LED_RED2_OFF;
        led_time = get_timerval();
        while ((signed char)(get_timerval() - led_time) < (SECOND_PERIOD / 2L / TICK_PERIOD)) {}
      }    
  }


//==============================================================================
//
// Section 6 Main
//
//==============================================================================

static void show_in_state(void)
  {
    #if (LED_MODE == STANDARD)
    switch(in_state)
      {
        case AC:    LED_RED1_OFF; LED_BLUE1_ON;   LED_BLUE2_ON;   LED_RED2_OFF;
            break;
        case DC_P:  LED_RED1_OFF; LED_BLUE1_OFF;  LED_BLUE2_OFF;  LED_RED2_ON;
            break;
        case DC_N:  LED_RED1_ON;  LED_BLUE1_OFF;  LED_BLUE2_OFF;  LED_RED2_OFF;
            break;
        case DCC_P: LED_RED1_ON;  LED_BLUE1_OFF;  LED_BLUE2_ON;   LED_RED2_OFF;
            break;
        case DCC_N: LED_RED1_OFF; LED_BLUE1_ON;   LED_BLUE2_OFF;  LED_RED2_ON;
            break;
      }
    #endif
  }

int main(void)
  {
    signed char my_tick;
    
    init_main();

    my_tick = get_timerval();
    set_sleep_mode(SLEEP_MODE_IDLE);   // SLEEP_MODE_PWR_DOWN , SLEEP_MODE_PWR_SAVE

    sei();

    // blink_red();                // say hello
    #if (LED_MODE == TEST)
      LED_RED1_OFF; LED_BLUE1_OFF;
    #endif

    while(1)
      {
        if (semaphor_get(C_DCC_Received))
          {                                                  // DCC is received, MCUCR has polarity
            my_tick = get_timerval();
            if (MCUCR & (1<<ISC00)) in_state = DCC_P;
            else in_state = DCC_N;
            show_in_state();
          }
        if ((signed char)(get_timerval()-my_tick) > AC_TIMEOUT_PERIOD / TICK_PERIOD)
          { //    // 100ms over
            my_tick = get_timerval();
            if (semaphor_get(C_DCC_Edge))
              {                                              // some Edges, must be AC or DCC
                                                             // when DCC, we had a DCC_Received
                in_state = AC;
              }
            else
              { // DC
                if (DCC_IN_HIGH) in_state = DC_P;
                else in_state = DC_N;
              }
            show_in_state();
          }
        sleep_cpu();    // wakeup with timer int
      }
    return(0);  // dummy  
  }
